////////////////////////////////////////////////////////////////////////////////
//
// Filename: 	wavegen.v
//
// Project:	Verilog Tutorial Example file
//
// Purpose:	Provide two bus ports to generate an outgoing waveform.  The
//		first port controls the waveform memory, the second port
//	controls the waveform reader: start/pause/reset, playback position,
//	and playback speed.
//
// Creator:	Dan Gisselquist, Ph.D.
//		Gisselquist Technology, LLC
//
////////////////////////////////////////////////////////////////////////////////
//
// Written and distributed by Gisselquist Technology, LLC
//
// This program is hereby granted to the public domain.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTIBILITY or
// FITNESS FOR A PARTICULAR PURPOSE.
//
////////////////////////////////////////////////////////////////////////////////
//
//
`default_nettype	none
//
module	wavegen(i_clk, i_reset,
		i_mem_cyc, i_mem_stb, i_mem_we, i_mem_addr, i_mem_data,
			i_mem_sel, o_mem_stall, o_mem_ack, o_mem_data,
		i_ctrl_cyc, i_ctrl_stb, i_ctrl_we, i_ctrl_addr, i_ctrl_data,
			i_ctrl_sel, o_ctrl_stall, o_ctrl_ack, o_ctrl_data,
		o_sample);
	parameter	LGNA = 5;	// Log(base two) size of our memory
	localparam	DW = 32; // Bus width
	//
	input	wire		i_clk, i_reset;
	//
	input	wire		i_mem_cyc, i_mem_stb, i_mem_we;
	input	wire [LGNA-3:0]	i_mem_addr;
	input	wire [DW-1:0]	i_mem_data;
	input	wire [DW/8-1:0]	i_mem_sel;
	output	reg		o_mem_stall;
	output	reg		o_mem_ack;
	output	reg [DW-1:0]	o_mem_data;
	//
	input	wire		i_ctrl_cyc, i_ctrl_stb, i_ctrl_we;
	input	wire	[1:0]	i_ctrl_addr;
	input	wire [DW-1:0]	i_ctrl_data;
	input	wire [DW/8-1:0]	i_ctrl_sel;
	output	reg		o_ctrl_stall;
	output	reg		o_ctrl_ack;
	output	reg [DW-1:0]	o_ctrl_data;
	//
	output	reg	[7:0]	o_sample;
	//

	reg	[DW-1:0]	mem [0:(1<<(LGNA-2))-1];

	reg		[31:0]	memaddr;
	reg		[31:0]	speed;
	reg			playing, wrap, reset_addr;


	////////////////////////////////////////////////////////////////////////
	//
	// The memory bus interface
	//

	// Writes
	always @(posedge i_clk)
	if (i_mem_stb && i_mem_we)
	begin
		if (i_mem_sel[3])
			mem[i_mem_addr][31:24] <= i_mem_data[31:24];
		if (i_mem_sel[2])
			mem[i_mem_addr][23:16] <= i_mem_data[23:16];
		if (i_mem_sel[3])
			mem[i_mem_addr][15: 8] <= i_mem_data[15: 8];
		if (i_mem_sel[2])
			mem[i_mem_addr][ 7: 0] <= i_mem_data[ 7: 0];
	end

	// Memory reads
	always @(posedge i_clk)
		o_mem_data <= mem[i_mem_addr];

	// Memory bus signaling
	always @(*)
		o_mem_stall <= 1'b0;

	always @(posedge i_clk)
		o_mem_ack <= i_mem_stb;

	////////////////////////////////////////////////////////////////////////
	//
	// The control bus interface
	//

	always @(posedge i_clk)
	if (i_ctrl_stb && i_ctrl_we && i_ctrl_addr == 0 && i_ctrl_sel[0])
	begin
		playing    <= i_ctrl_data[0];
		reset_addr <= i_ctrl_data[1];
		wrap       <= i_ctrl_data[2];
	end else
		reset_addr <= 1'b0;

	always @(posedge i_clk)
	if (i_ctrl_stb && i_ctrl_we && i_ctrl_addr == 1)
		memaddr <= i_ctrl_data;
	else if (reset_addr)
		memaddr <= 0;
	else if (playing)
		memaddr <= memaddr + speed;

	always @(posedge i_clk)
	if (i_ctrl_stb && i_ctrl_we && i_ctrl_addr == 2)
		speed <= i_ctrl_data;

	wire	[DW-1:0]	w_control_data;

	assign	w_control_data = { 29'h0, wrap, 1'b0, playing };

	always @(posedge i_clk)
	case(i_ctrl_addr)
	0: o_ctrl_data <= w_control_data;
	1: o_ctrl_data <= memaddr;
	2: o_ctrl_data <= speed;
	3: o_ctrl_data <= 0;
	endcase

	always @(*)
		o_ctrl_stall = 1'b0;

	always @(posedge i_clk)
		o_ctrl_ack = i_ctrl_stb;

	////////////////////////////////////////////////////////////////////////
	//
	// Output logic
	//
	always @(posedge i_clk)
		memword <= mem[memaddr[31:30-LGNA]];

	always @(posedge i_clk)
		subaddr <= memaddr[29-LGNA:32-LGNA];

	always @(posedge i_clk)
		o_sample <= memword >> (subaddr * 8);


	// Make verilator happy
	// Verilator lint_off UNUSED
	wire	unused;
	assign	unused = &{ 1'b0, i_wb_cyc, i_wb_sel, scaled[32], scaled[15:0] };
	// Verilator lint_on  UNUSED

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
// Formal properties section
//
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
`ifdef	FORMAL
	localparam	F_LGDEPTH=2;

	reg		f_past_valid;
	wire [F_LGDEPTH-1:0] fmem_nreqs, fmem_nacks, fmem_outstanding;
	wire [F_LGDEPTH-1:0] fctrl_nreqs, fctrl_nacks, fctrl_outstanding;

	////////////////////////////////////////////////////////////////////////
	//
	// Assume an initial reset
	//

	initial	f_past_valid = 1'b0;
	always @(posedge i_clk)
		f_past_valid <= 1'b1;

	always @(*)
	if (!f_past_valid)
		assume(i_reset);

	////////////////////////////////////////////////////////////////////////
	//
	// Verify our bus interfaces
	//

	fwb_slave #(.AW(LGNA-2),
		.F_LGDEPTH(F_LGDEPTH),
		.F_MAX_STALL(1),
		.F_MAX_ACK_DELAY(2))
	fmem (i_clk, i_reset,
		i_mem_cyc, i_mem_stb, i_mem_we, i_mem_addr,
	    	i_mem_data, i_mem_sel,
		o_mem_ack, o_mem_stall, o_mem_data, 1'b0,
		fmem_nreqs, fmem_nacks, fmem_outstanding);

	always @(*)
	if (i_mem_cyc)
		assert(fmem_outstanding == (o_mem_ack ? 1:0));

	fwb_slave #(.AW(2),
		.F_LGDEPTH(F_LGDEPTH),
		.F_MAX_STALL(1),
		.F_MAX_ACK_DELAY(2))
	fctrl (i_clk, i_reset,
		i_wb_cyc, i_wb_stb, i_wb_we, i_wb_addr,
	    	i_wb_data, i_wb_sel,
		o_wb_ack, o_wb_stall, o_wb_data, 1'b0,
		fctrl_nreqs, fctrl_nacks, fctrl_outstanding);

	always @(*)
	if (i_ctrl_cyc)
		assert(fctrl_outstanding == (o_ctrl_ack ? 1:0));

	////////////////////////////////////////////////////////////////////////
	//
	// Check that we got our memory interface right
	//
	(* anyconst *) reg	[DW-1:0]	f_addr;
	reg	[DW-1:0]	f_data;

	always @(*)
	if (!f_past_valid)
		assume(f_data == mem[f_addr]);

	always @(posedge i_clk)
	if (i_mem_stb && i_mem_we && i_mem_addr == f_addr)
		f_data <= i_mem_addr;

	always @(posedge i_clk)
	if (f_past_valid && !$past(i_reset)
		&& $past(i_mem_stb && i_mem_addr == f_addr))
	begin
		assert(o_mem_data == f_data);
		assert(o_mem_ack);
	end

	////////////////////////////////////////////////////////////////////////
	//
	// Check that we got our control interface right
	//

	// ... your turn.  What would verify this?


	////////////////////////////////////////////////////////////////////////
	//
	// Check that we have the correct output
	//

	// ... your turn.  What would verify this?

	////////////////////////////////////////////////////////////////////////
	//
	// cover() ... let's see what kind of throughput we can achieve
	//
	reg	cvr_mem_returns, cvr_ctrl_returns;

	always @(posedge i_clk)
	if (i_reset)
		cvr_mem_returns = 0;
	else if (o_mem_ack)
		cvr_mem_returns <= cvr_mem_returns + 1;

	always @(*)
		cover(cvr_mem_returns == 4);

	always @(posedge i_clk)
	if (i_reset)
		cvr_ctrl_returns = 0;
	else if (o_mem_ack)
		cvr_ctrl_returns <= cvr_ctrl_returns + 1;

	always @(*)
		cover(cvr_ctrl_returns == 4);
`endif
endmodule
